
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "nssl.h"
#include "sock.h"
#include "utils.h"
#include "sslut.h"
#include "pem_bin.h"



/* initialization for a SSL connect() */

int	ssl_init_connect (SSL **ssl_socket, SSL_CTX **ctx)
{
	SSLeay_add_ssl_algorithms ();

	if((*ctx = SSL_CTX_new (SSLv3_client_method())) == NULL)
	   {
		fprintf (stderr, "could not initialize SSL_CTX structure\n");
		return (-1);
	   }
  
	SSL_CTX_set_quiet_shutdown (*ctx, 1);
 
	*ssl_socket = SSL_new (*ctx);

	return (0);
}



/* initialization for a SSL accept() */

int	ssl_init_accept (SSL_CTX **ctx, char *pemfile)
{
	SSLeay_add_ssl_algorithms ();

	if((*ctx = SSL_CTX_new (SSLv23_server_method())) == NULL)
	   {
		fprintf(stderr, "could not initialize SSL_CTX structure\n");
		return (-1);
	   }
  
	SSL_CTX_set_quiet_shutdown (*ctx,1);

	SSL_CTX_set_session_cache_mode (*ctx,SSL_SESS_CACHE_OFF);

	if (pemfile)
	   {
		if (!SSL_CTX_use_RSAPrivateKey_file (*ctx, pemfile, SSL_FILETYPE_PEM))
		   {
			fprintf(stderr, "could not load RSA private key from [%s]\n", pemfile);
			return (-1);
		   }

		if (!SSL_CTX_use_certificate_file (*ctx, pemfile, SSL_FILETYPE_PEM))
		   {
			fprintf(stderr, "could not load certificate from [%s]\n", pemfile);
			return (-1);
		   }
	   }
	else
	   {
		if (!SSL_CTX_use_RSAPrivateKey_ASN1 (*ctx, priv_bin, priv_bin_len))
		   {
			fprintf(stderr, "could not load hardcoded RSA private key\n");
			return (-1);
		   }

		if (!SSL_CTX_use_certificate_ASN1 (*ctx, cert_bin_len, cert_bin))
		   {
			fprintf(stderr, "could not load certificate\n");
			return (-1);
		   }
	   }

        if (!SSL_CTX_check_private_key (*ctx))
           {
		fprintf(stderr, "private key from [%s] is not valid\n", pemfile);
		return (-1);
           }

	return (0);
}



/* connect to an SSL port and return socket (ssl_socket and ssl_ctx are initialized) */

int	ssl_connect_ip (struct in_addr *ip, 
			unsigned short port, unsigned short sport,
                        SSL **ssl_socket, SSL_CTX **ctx)
{
	int sock;

	if (ssl_init_connect (ssl_socket, ctx) < 0)
	   {
		return (-1);
	   }
  
	if ((sock = connect_tcp (ip, port, sport)) == -1) 
	   {
		ssl_close_connection (*ssl_socket, *ctx);
		return (-1);
	   }
  
	if (ssl_handshake_timeout (*ssl_socket, sock, SSL_HANDSHAKE_TIMEOUT) < 0)
	   {
		ssl_close_all (sock, *ssl_socket, *ctx);     
		return (-1);
	   }

	return (sock);
}



/* set the ssl socket into accept mode and wait for an incoming connection */

int	ssl_wait_for_connection (int sock_accept, SSL **ssl_socket, SSL_CTX *ctx)
{
	int	sock, ret;

        sock = wait_for_connect (sock_accept);

        *ssl_socket = SSL_new (ctx);

        SSL_set_fd (*ssl_socket, sock);

        SSL_set_accept_state (*ssl_socket);

        ret = SSL_accept (*ssl_socket);

	if (ssl_check_retcode (*ssl_socket, ret) == -1)
	   {
		return (-1);
	   }

	return (sock);
}



/* check error code */

int	ssl_check_retcode (SSL *ssl_socket, int ret)
{
	switch (SSL_get_error (ssl_socket,ret))
           {
		case SSL_ERROR_NONE:
			break;
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_X509_LOOKUP:
		case SSL_ERROR_WANT_CONNECT:
			fprintf (stderr, "ssl error: should reconnect\n");
			break;
		case SSL_ERROR_SYSCALL:
			fprintf (stderr, "ssl error: syscall intr\n");
			return (-1);
		case SSL_ERROR_SSL:
			fprintf (stderr, "ssl error occured\n");
			return (-1);
		case SSL_ERROR_ZERO_RETURN:
			fprintf (stderr, "ssl error: zero return\n");
			return (-1);
           }

	return (0);
}



/*
 * perform a SSL handshake with a timeout so that a silent port
 * will not block the connection forever.
 */

int	ssl_handshake_timeout (SSL *ssl_socket, int socket, int timeout)
{
	int oldflags = 0, ssl_ret;

	SSL_set_fd (ssl_socket, socket);

	SSL_set_connect_state (ssl_socket);  
  
	if ( (oldflags = set_nonblocking_mode (socket)) < 0) 
	   {
		return (-1);
	   }

	/* on local connections, SSL_do_handshake may finish */
	ssl_ret = SSL_do_handshake(ssl_socket);

	if ( (unset_nonblocking_mode (socket, oldflags)) < 0) 
	   {
		return (-1);
	   }

	/* on remote connections, wait for handshake answer and finish ssl connect */
	if (!SSL_is_init_finished (ssl_socket))
	   {
		if (data_available (socket, timeout) < 0)
		   {
			return (-1);
		   }

		ssl_ret = SSL_connect(ssl_socket);
	   }

	switch (ssl_ret)
	   {
		default:
		case (-1): 
			fprintf (stderr, "TLS/SSL handshake failed.\n");
			return (-1);
		case (0):
			fprintf (stderr, "TLS/SSL handshake failed cleanly.\n");
			return (-1);
		case (1):
			return (0);
	   }
}



/* send raw data */

int	ssl_write (SSL* ssl_sock, char *str, unsigned int len)
{
	return ( SSL_write (ssl_sock, str, len) );
}



/* write a null terminated string */

int	ssl_writes (SSL* ssl_sock, char *str)
{
	return ( SSL_write (ssl_sock, str, strlen(str)) );
}



/* terminate SSL connection */ 
 
int	ssl_close_connection (SSL *ssl_socket, SSL_CTX *ctx)
{
	if (ssl_socket) 
	   {
		SSL_shutdown (ssl_socket);
		SSL_free (ssl_socket);
	   }

	if (ctx)
	   {
		SSL_CTX_free (ctx);
	   }

	return (0);
}
 
 

/* close all (SSL socket and file descriptor) */

int	ssl_close_all (int fd, SSL *ssl_socket, SSL_CTX *ctx)
{
	ssl_close_connection(ssl_socket, ctx);

	if (fd)
	   {
		close(fd);
	   }

	return (0);
}
 


int     ssl_multiplex_loop (SSL *ssl_socket, int sockin, int sockprint, int sockout)
{
        int                     num = 0, fdhigh, inclose = 0;
        unsigned char           buf[4096];
        fd_set                  fdread;

        fdhigh = (sockin <= sockout) ? sockout : sockin;

        while (1)
           {
                if (SSL_pending (ssl_socket))
		   {
                        num = SSL_read (ssl_socket, buf, sizeof (buf));
                        if (num <= 0) return -1;
                        num = write (sockin, buf, num);
		   }
 

                FD_ZERO (&fdread);
                FD_SET  (sockout, &fdread);
                if (!inclose)
                   {
                        FD_SET  (sockin,  &fdread);
                        fdhigh = (sockin <= sockout) ? sockout : sockin;
                   }
                else
                   {
                        fdhigh = sockout;
                   }

                select (fdhigh+1, &fdread, NULL, NULL, NULL);

                if ( FD_ISSET (sockout, &fdread) )
                   {
                        num = SSL_read (ssl_socket, buf, sizeof (buf));
                        if (num <= 0) return -1;
                        if (f_hexdump) hexdump (DATA_INPUT, buf, num);
                        num = write (sockprint, buf, num);
                        if (num <= 0) return -1;
                   }

                if ( FD_ISSET (sockin, &fdread) )
                   {
                        num = read (sockin, buf, sizeof (buf));
                        if (num <= 0)
                           {
                                if (sockin == STDIN_FILENO)
                                   {
                                        inclose = 1;
                                   }
                                else
                                   {
                                        return -1;
                                   }
                           }
                        else
                           {
                                if (f_hexdump) hexdump (DATA_OUTPUT, buf, num);
				num = SSL_write (ssl_socket, buf, num);
                                if (num <= 0) return -1;
                           }

                   }

           }

        return 0;
}



/* print current cipher used */

void	ssl_print_cipher (SSL *ssl_socket)
{
	fprintf (stderr, "using SSL cipher [%s]\n", SSL_get_cipher (ssl_socket));
}



int     ssl_read_banner (SSL *ssl_socket, int socket)
{
        int     num;
        char    buf[2048];

        if (SSL_pending (ssl_socket))
           {
              num = SSL_read (ssl_socket, buf, sizeof (buf));
           }
	else
	   {
		if (data_available (socket, BANNER_TIMEOUT) < 0)
		   {
			return (-1);
		   }
                num = SSL_read (ssl_socket, buf, sizeof (buf));
           }

        if (num > 0)
           {
                if (f_hexdump) hexdump (DATA_INPUT, buf, num);
                write (STDOUT_FILENO, buf, num);
                return (0);
           }
        else
           {
                return (-1);
           }
}


